# Copyright 2024 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import os
import stat
import subprocess
from unittest import mock

from octavia_lib.common import constants as lib_consts
from webob import exc

from octavia.amphorae.backends.utils import nftable_utils
from octavia.common import constants as consts
from octavia.common import exceptions
import octavia.tests.unit.base as base


class TestNFTableUtils(base.TestCase):
    @mock.patch('os.open')
    @mock.patch('os.path.isfile')
    def test_write_nftable_rules_file_exists(self, mock_isfile, mock_open):
        """Test when a rules file exists and no new rules

        When an existing rules file is present and we call
        write_nftable_rules_file with no rules, the method should not
        overwrite the existing rules.
        """
        mock_isfile.return_value = True

        nftable_utils.write_nftable_rules_file('fake-eth2', [])

        mock_open.assert_not_called()

    @mock.patch('os.open')
    @mock.patch('os.path.isfile')
    def test_write_nftable_rules_file_rules(self, mock_isfile, mock_open):
        """Test when a rules file exists and rules are passed in

        This should create a simple rules file with the base chain and rules.
        """
        mock_isfile.return_value = True
        mock_open.return_value = 'fake-fd'

        test_rule_1 = {consts.CIDR: None,
                       consts.PROTOCOL: lib_consts.PROTOCOL_TCP,
                       consts.PORT: 1234}
        test_rule_2 = {consts.CIDR: '192.0.2.0/24',
                       consts.PROTOCOL: consts.VRRP,
                       consts.PORT: 4321}

        mocked_open = mock.mock_open()
        with mock.patch.object(os, 'fdopen', mocked_open):
            nftable_utils.write_nftable_rules_file(
                'fake-eth2', [test_rule_1, test_rule_2])

        mocked_open.assert_called_once_with('fake-fd', 'w')
        mock_open.assert_called_once_with(
            consts.NFT_RULES_FILE,
            (os.O_WRONLY | os.O_CREAT | os.O_TRUNC),
            (stat.S_IRUSR | stat.S_IWUSR))

        handle = mocked_open()
        handle.write.assert_has_calls([
            mock.call(f'table {consts.NFT_FAMILY} {consts.NFT_TABLE} '
                      '{}\n'),
            mock.call(f'delete table {consts.NFT_FAMILY} '
                      f'{consts.NFT_TABLE}\n'),
            mock.call(f'table {consts.NFT_FAMILY} {consts.NFT_TABLE} '
                      '{\n'),
            mock.call(f'  chain {consts.NFT_CHAIN} {{\n'),
            mock.call('    type filter hook input priority filter; '
                      'policy drop;\n'),
            mock.call('      ct state vmap { established : accept, related : '
                      'accept, invalid : drop }\n'),
            mock.call('      iif lo accept\n'),
            mock.call('      ip saddr 127.0.0.0/8 drop\n'),
            mock.call('      ip6 saddr ::1 drop\n'),
            mock.call('      icmp type destination-unreachable accept\n'),
            mock.call('      icmpv6 type { nd-neighbor-solicit, '
                      'nd-router-advert, nd-neighbor-advert, packet-too-big, '
                      'destination-unreachable } accept\n'),
            mock.call('      udp sport 67 udp dport 68 accept\n'),
            mock.call('      udp sport 547 udp dport 546 accept\n'),
            mock.call('      iifname eth1 goto amphora_vip_chain\n'),
            mock.call('  }\n'),
            mock.call('  chain amphora_vip_chain {\n'),
            mock.call('      tcp dport 1234 accept\n'),
            mock.call('      ip saddr 192.0.2.0/24 ip protocol 112 accept\n'),
            mock.call('  }\n'),
            mock.call('}\n')
        ])

    @mock.patch('os.open')
    @mock.patch('os.path.isfile')
    def test_write_nftable_rules_file_missing(self, mock_isfile, mock_open):
        """Test when a rules file does not exist and no new rules

        This should create a simple rules file with the base chain.
        """
        mock_isfile.return_value = False
        mock_open.return_value = 'fake-fd'

        mocked_open = mock.mock_open()
        with mock.patch.object(os, 'fdopen', mocked_open):
            nftable_utils.write_nftable_rules_file('fake-eth2', [])

        mocked_open.assert_called_once_with('fake-fd', 'w')
        mock_open.assert_called_once_with(
            consts.NFT_RULES_FILE,
            (os.O_WRONLY | os.O_CREAT | os.O_TRUNC),
            (stat.S_IRUSR | stat.S_IWUSR))

        handle = mocked_open()
        handle.write.assert_has_calls([
            mock.call(f'table {consts.NFT_FAMILY} {consts.NFT_TABLE} '
                      '{\n'),
            mock.call(f'  chain {consts.NFT_CHAIN} {{\n'),
            mock.call('    type filter hook input priority filter; '
                      'policy drop;\n'),
            mock.call('      icmp type destination-unreachable accept\n'),
            mock.call('      icmpv6 type { nd-neighbor-solicit, '
                      'nd-router-advert, nd-neighbor-advert, packet-too-big, '
                      'destination-unreachable } accept\n'),
            mock.call('      udp sport 67 udp dport 68 accept\n'),
            mock.call('      udp sport 547 udp dport 546 accept\n'),
            mock.call('  }\n'),
            mock.call('}\n')
        ])

    @mock.patch('octavia.common.utils.ip_version')
    def test__build_rule_cmd(self, mock_ip_version):

        mock_ip_version.side_effect = [4, 6, 99]

        cmd = nftable_utils._build_rule_cmd({
            consts.CIDR: '192.0.2.0/24',
            consts.PROTOCOL: lib_consts.PROTOCOL_SCTP,
            consts.PORT: 1234})
        self.assertEqual('ip saddr 192.0.2.0/24 sctp dport 1234 accept', cmd)

        cmd = nftable_utils._build_rule_cmd({
            consts.CIDR: '2001:db8::/32',
            consts.PROTOCOL: lib_consts.PROTOCOL_TCP,
            consts.PORT: 1235})
        self.assertEqual('ip6 saddr 2001:db8::/32 tcp dport 1235 accept', cmd)

        self.assertRaises(exc.HTTPBadRequest, nftable_utils._build_rule_cmd,
                          {consts.CIDR: '192/32',
                           consts.PROTOCOL: lib_consts.PROTOCOL_TCP,
                           consts.PORT: 1237})

        cmd = nftable_utils._build_rule_cmd({
            consts.CIDR: None,
            consts.PROTOCOL: lib_consts.PROTOCOL_UDP,
            consts.PORT: 1236})
        self.assertEqual('udp dport 1236 accept', cmd)

        cmd = nftable_utils._build_rule_cmd({
            consts.CIDR: None,
            consts.PROTOCOL: consts.VRRP,
            consts.PORT: 1237})
        self.assertEqual('ip protocol 112 accept', cmd)

        self.assertRaises(exc.HTTPBadRequest, nftable_utils._build_rule_cmd,
                          {consts.CIDR: None,
                           consts.PROTOCOL: 'bad-protocol',
                           consts.PORT: 1237})

    @mock.patch('octavia.amphorae.backends.utils.network_namespace.'
                'NetworkNamespace')
    @mock.patch('subprocess.check_output')
    def test_load_nftables_file(self, mock_check_output, mock_netns):

        mock_netns.side_effect = [
            mock.DEFAULT,
            subprocess.CalledProcessError(cmd=consts.NFT_CMD, returncode=-1),
            exceptions.AmphoraNetworkConfigException]

        nftable_utils.load_nftables_file()

        mock_netns.assert_called_once_with(consts.AMPHORA_NAMESPACE)
        mock_check_output.assert_called_once_with([
            consts.NFT_CMD, '-o', '-f', consts.NFT_RULES_FILE],
            stderr=subprocess.STDOUT)

        self.assertRaises(subprocess.CalledProcessError,
                          nftable_utils.load_nftables_file)

        self.assertRaises(exceptions.AmphoraNetworkConfigException,
                          nftable_utils.load_nftables_file)



# Copyright Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Add FAILOVER_STOPPED to provisioning_status table

Revision ID: 0995c26fc506
Revises: 31f7653ded67
Create Date: 2022-03-24 04:53:10.768658

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '0995c26fc506'
down_revision = '31f7653ded67'


def upgrade():
    insert_table = sa.sql.table(
        'provisioning_status',
        sa.sql.column('name', sa.String),
        sa.sql.column('description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'FAILOVER_STOPPED'},
        ]
    )



# Copyright 2015 Rackspace
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os
from unittest import mock

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture
from werkzeug import exceptions as wz_exceptions

from octavia.amphorae.backends.agent.api_server import osutils
from octavia.amphorae.backends.agent.api_server import plug
from octavia.common import constants
import octavia.tests.unit.base as base

FAKE_CIDR_IPV4 = '10.0.0.0/24'
FAKE_GATEWAY_IPV4 = '10.0.0.1'
FAKE_IP_IPV4 = '10.0.0.2'
FAKE_CIDR_IPV6 = '2001:db8::/32'
FAKE_GATEWAY_IPV6 = '2001:db8::1'
FAKE_IP_IPV6 = '2001:db8::2'
FAKE_IP_IPV6_EXPANDED = '2001:0db8:0000:0000:0000:0000:0000:0002'
FAKE_MAC_ADDRESS = 'ab:cd:ef:00:ff:22'
FAKE_INTERFACE = 'eth33'


class TestPlug(base.TestCase):
    def setUp(self):
        super().setUp()
        self.mock_platform = mock.patch("distro.id").start()
        self.mock_platform.return_value = "ubuntu"
        self.osutil = osutils.BaseOS.get_os_util()
        self.test_plug = plug.Plug(self.osutil)
        self.addCleanup(self.mock_platform.stop)

    @mock.patch('pyroute2.IPRoute', create=True)
    def test__interface_by_mac_case_insensitive_ubuntu(self, mock_ipr):
        mock_ipr_instance = mock.MagicMock()
        mock_ipr_instance.link_lookup.return_value = [33]
        mock_ipr_instance.get_links.return_value = ({
            'attrs': [('IFLA_IFNAME', FAKE_INTERFACE)]},)
        mock_ipr().__enter__.return_value = mock_ipr_instance

        interface = self.test_plug._interface_by_mac(FAKE_MAC_ADDRESS.upper())
        self.assertEqual(FAKE_INTERFACE, interface)
        mock_ipr_instance.get_links.assert_called_once_with(33)

    @mock.patch('pyroute2.IPRoute', create=True)
    def test__interface_by_mac_not_found(self, mock_ipr):
        mock_ipr_instance = mock.MagicMock()
        mock_ipr_instance.link_lookup.return_value = []
        mock_ipr().__enter__.return_value = mock_ipr_instance

        fd_mock = mock.mock_open()
        open_mock = mock.Mock()
        isfile_mock = mock.Mock()
        with mock.patch('os.open', open_mock), mock.patch.object(
                os, 'fdopen', fd_mock), mock.patch.object(
                os.path, 'isfile', isfile_mock):
            self.assertRaises(wz_exceptions.HTTPException,
                              self.test_plug._interface_by_mac,
                              FAKE_MAC_ADDRESS.upper())
        open_mock.assert_called_once_with('/sys/bus/pci/rescan', os.O_WRONLY)
        fd_mock().write.assert_called_once_with('1')

    @mock.patch('pyroute2.IPRoute', create=True)
    def test__interface_by_mac_case_insensitive_rh(self, mock_ipr):
        mock_ipr_instance = mock.MagicMock()
        mock_ipr_instance.link_lookup.return_value = [33]
        mock_ipr_instance.get_links.return_value = ({
            'attrs': [('IFLA_IFNAME', FAKE_INTERFACE)]},)
        mock_ipr().__enter__.return_value = mock_ipr_instance

        with mock.patch('distro.id', return_value='centos'):
            osutil = osutils.BaseOS.get_os_util()
            self.test_plug = plug.Plug(osutil)
            interface = self.test_plug._interface_by_mac(
                FAKE_MAC_ADDRESS.upper())
            self.assertEqual(FAKE_INTERFACE, interface)
            mock_ipr_instance.get_links.assert_called_once_with(33)

    @mock.patch('octavia.amphorae.backends.agent.api_server.plug.Plug.'
                '_interface_by_mac', return_value=FAKE_INTERFACE)
    @mock.patch('pyroute2.NSPopen', create=True)
    @mock.patch.object(plug, "webob")
    @mock.patch('pyroute2.IPRoute', create=True)
    @mock.patch('pyroute2.NetNS', create=True)
    @mock.patch('subprocess.check_output')
    @mock.patch('shutil.copytree')
    @mock.patch('os.makedirs')
    def test_plug_vip_ipv4(self, mock_makedirs, mock_copytree,
                           mock_check_output, mock_netns, mock_pyroute2,
                           mock_webob, mock_nspopen, mock_by_mac):
        m = mock.mock_open()
        with mock.patch('os.open'), mock.patch.object(os, 'fdopen', m):
            self.test_plug.plug_vip(
                vip=FAKE_IP_IPV4,
                subnet_cidr=FAKE_CIDR_IPV4,
                gateway=FAKE_GATEWAY_IPV4,
                mac_address=FAKE_MAC_ADDRESS
            )
        mock_webob.Response.assert_any_call(json={
            'message': 'OK',
            'details': f'VIPs plugged on interface eth1: {FAKE_IP_IPV4}'
        }, status=202)

    @mock.patch('octavia.amphorae.backends.agent.api_server.plug.Plug.'
                '_interface_by_mac', return_value=FAKE_INTERFACE)
    @mock.patch('pyroute2.NSPopen', create=True)
    @mock.patch.object(plug, "webob")
    @mock.patch('pyroute2.IPRoute', create=True)
    @mock.patch('pyroute2.NetNS', create=True)
    @mock.patch('subprocess.check_output')
    @mock.patch('shutil.copytree')
    @mock.patch('os.makedirs')
    def test_plug_vip_ipv6(self, mock_makedirs, mock_copytree,
                           mock_check_output, mock_netns, mock_pyroute2,
                           mock_webob, mock_nspopen, mock_by_mac):
        conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        conf.config(group='controller_worker',
                    loadbalancer_topology=constants.TOPOLOGY_ACTIVE_STANDBY)
        m = mock.mock_open()
        with mock.patch('os.open'), mock.patch.object(os, 'fdopen', m):
            self.test_plug.plug_vip(
                vip=FAKE_IP_IPV6,
                subnet_cidr=FAKE_CIDR_IPV6,
                gateway=FAKE_GATEWAY_IPV6,
                mac_address=FAKE_MAC_ADDRESS
            )
        mock_webob.Response.assert_any_call(json={
            'message': 'OK',
            'details': f'VIPs plugged on interface eth1: '
                       f'{FAKE_IP_IPV6_EXPANDED}'
        }, status=202)

    @mock.patch('octavia.amphorae.backends.agent.api_server.plug.Plug.'
                '_interface_by_mac', return_value=FAKE_INTERFACE)
    @mock.patch('pyroute2.NSPopen', create=True)
    @mock.patch.object(plug, "webob")
    @mock.patch('pyroute2.IPRoute', create=True)
    @mock.patch('pyroute2.NetNS', create=True)
    @mock.patch('subprocess.check_output')
    @mock.patch('shutil.copytree')
    @mock.patch('os.makedirs')
    def test_plug_vip_ipv4_and_ipv6(
            self, mock_makedirs, mock_copytree,
            mock_check_output, mock_netns,
            mock_pyroute2, mock_webob, mock_nspopen, mock_by_mac):
        conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        conf.config(group='controller_worker',
                    loadbalancer_topology=constants.TOPOLOGY_ACTIVE_STANDBY)
        additional_vips = [
            {'ip_address': FAKE_IP_IPV4, 'subnet_cidr': FAKE_CIDR_IPV4,
             'host_routes': [], 'gateway': FAKE_GATEWAY_IPV4}
        ]
        m = mock.mock_open()
        with mock.patch('os.open'), mock.patch.object(os, 'fdopen', m):
            self.test_plug.plug_vip(
                vip=FAKE_IP_IPV6,
                subnet_cidr=FAKE_CIDR_IPV6,
                gateway=FAKE_GATEWAY_IPV6,
                mac_address=FAKE_MAC_ADDRESS,
                additional_vips=additional_vips
            )
        mock_webob.Response.assert_any_call(json={
            'message': 'OK',
            'details': 'VIPs plugged on interface {interface}: {vips}'.format(
                vips=", ".join([FAKE_IP_IPV6_EXPANDED, FAKE_IP_IPV4]),
                interface='eth1')
        }, status=202)

    @mock.patch.object(plug, "webob")
    @mock.patch('pyroute2.IPRoute', create=True)
    @mock.patch('pyroute2.NetNS', create=True)
    @mock.patch('subprocess.check_output')
    @mock.patch('shutil.copytree')
    @mock.patch('os.makedirs')
    def test_plug_vip_bad_ip(self, mock_makedirs, mock_copytree,
                             mock_check_output, mock_netns, mock_pyroute2,
                             mock_webob):
        m = mock.mock_open()
        BAD_IP_ADDRESS = "error"
        with mock.patch('os.open'), mock.patch.object(os, 'fdopen', m):
            self.test_plug.plug_vip(
                vip=BAD_IP_ADDRESS,
                subnet_cidr=FAKE_CIDR_IPV4,
                gateway=FAKE_GATEWAY_IPV4,
                mac_address=FAKE_MAC_ADDRESS
            )
        mock_webob.Response.assert_any_call(
            json={'message': ("Invalid VIP: '{ip}' does not appear to be an "
                              "IPv4 or IPv6 address").format(
                                  ip=BAD_IP_ADDRESS)},
            status=400)

    @mock.patch.object(plug, "webob")
    @mock.patch('pyroute2.IPRoute', create=True)
    @mock.patch('pyroute2.NetNS', create=True)
    @mock.patch('subprocess.check_output')
    @mock.patch('shutil.copytree')
    @mock.patch('os.makedirs')
    def test_plug_vip_bad_vrrp_ip(self, mock_makedirs, mock_copytree,
                                  mock_check_output, mock_netns, mock_pyroute2,
                                  mock_webob):
        m = mock.mock_open()
        BAD_IP_ADDRESS = "error"
        with mock.patch('os.open'), mock.patch.object(os, 'fdopen', m):
            self.test_plug.plug_vip(
                vip=FAKE_IP_IPV4,
                subnet_cidr=FAKE_CIDR_IPV4,
                gateway=FAKE_GATEWAY_IPV4,
                mac_address=FAKE_MAC_ADDRESS,
                vrrp_ip=BAD_IP_ADDRESS
            )
        mock_webob.Response.assert_any_call(
            json={'message': ("Invalid VRRP Address: '{ip}' does not appear "
                              "to be an IPv4 or IPv6 address").format(
                                  ip=BAD_IP_ADDRESS)},
            status=400)

    @mock.patch("octavia.amphorae.backends.agent.api_server.osutils."
                "BaseOS.write_interface_file")
    def test_plug_lo(self, mock_write_interface):
        m = mock.mock_open()
        with mock.patch('os.open'), mock.patch.object(os, 'fdopen', m):
            self.test_plug.plug_lo()
        mock_write_interface.assert_called_once_with(interface='lo',
                                                     ip_address='127.0.0.1',
                                                     prefixlen=8)

    @mock.patch('pyroute2.NetNS', create=True)
    def test__netns_interface_exists(self, mock_netns):

        netns_handle = mock_netns.return_value.__enter__.return_value

        netns_handle.get_links.return_value = [{
            'attrs': [['IFLA_ADDRESS', '123'],
                      ['IFLA_IFNAME', 'eth0']]}]

        # Interface is found in netns
        self.assertTrue(self.test_plug._netns_interface_exists('123'))

        # Interface is not found in netns
        self.assertFalse(self.test_plug._netns_interface_exists('321'))

    @mock.patch.object(plug, "webob")
    @mock.patch('octavia.amphorae.backends.agent.api_server.plug.Plug.'
                '_netns_interface_exists', return_value=False)
    @mock.patch('octavia.amphorae.backends.agent.api_server.plug.Plug.'
                '_interface_by_mac', return_value=FAKE_INTERFACE)
    @mock.patch('pyroute2.IPRoute', create=True)
    @mock.patch('pyroute2.NetNS', create=True)
    @mock.patch("octavia.amphorae.backends.agent.api_server.osutils."
                "BaseOS.write_port_interface_file")
    @mock.patch("octavia.amphorae.backends.agent.api_server.osutils."
                "BaseOS.bring_interface_up")
    @mock.patch("octavia.amphorae.backends.agent.api_server.util."
                "send_member_advertisements")
    def test_plug_network(self, mock_send_member_adv,
                          mock_if_up, mock_write_port_interface,
                          mock_netns, mock_iproute,
                          mock_by_mac, mock_interface_exists, mock_webob):
        fixed_ips = [
            {'ip_address': FAKE_IP_IPV4,
             'subnet_cidr': FAKE_CIDR_IPV4,
             'gateway': FAKE_GATEWAY_IPV4,
             'host_routes': [
                 {'destination': '192.0.2.0/24',
                  'nexthop': '192.0.2.254'}]
             }]
        mtu = 1400
        m = mock.mock_open()
        with mock.patch('os.open'), mock.patch.object(os, 'fdopen', m):
            self.test_plug.plug_network(FAKE_MAC_ADDRESS, fixed_ips, 1400)

        mock_write_port_interface.assert_called_once_with(
            interface='eth2', fixed_ips=fixed_ips, mtu=mtu, is_sriov=False)
        mock_if_up.assert_called_once_with('eth2', 'network')
        mock_send_member_adv.assert_called_once_with(fixed_ips)

        mock_webob.Response.assert_any_call(
            json={'message': 'OK',
                  'details': 'Plugged on interface eth2'},
            status=202)

    @mock.patch.object(plug, "webob")
    @mock.patch('octavia.amphorae.backends.agent.api_server.plug.Plug.'
                '_netns_interface_exists', return_value=True)
    @mock.patch('octavia.amphorae.backends.agent.api_server.plug.Plug.'
                '_netns_interface_by_mac', return_value=FAKE_INTERFACE)
    @mock.patch('pyroute2.NetNS', create=True)
    @mock.patch("octavia.amphorae.backends.agent.api_server.osutils."
                "BaseOS.write_port_interface_file")
    @mock.patch("octavia.amphorae.backends.agent.api_server.osutils."
                "BaseOS.bring_interface_up")
    @mock.patch("octavia.amphorae.backends.agent.api_server.util."
                "send_member_advertisements")
    def test_plug_network_existing_interface(self, mock_send_member_adv,
                                             mock_if_up,
                                             mock_write_port_interface,
                                             mock_netns, mock_by_mac,
                                             mock_interface_exists,
                                             mock_webob):
        fixed_ips = [
            {'ip_address': FAKE_IP_IPV4,
             'subnet_cidr': FAKE_CIDR_IPV4,
             'gateway': FAKE_GATEWAY_IPV4,
             'host_routes': [
                 {'destination': '192.0.2.0/24',
                  'nexthop': '192.0.2.254'}]
             }, {'ip_address': FAKE_IP_IPV6,
                 'subnet_cidr': FAKE_CIDR_IPV6,
                 'gateway': FAKE_GATEWAY_IPV6,
                 'host_routes': [
                     {'destination': '2001:db8::/64',
                      'nexthop': '2001:db8::ffff'}]
                 }]
        mtu = 1400
        m = mock.mock_open()
        with mock.patch('os.open'), mock.patch.object(os, 'fdopen', m):
            self.test_plug.plug_network(FAKE_MAC_ADDRESS, fixed_ips, 1400)

        mock_write_port_interface.assert_called_once_with(
            interface=FAKE_INTERFACE, fixed_ips=fixed_ips, mtu=mtu,
            is_sriov=False)
        mock_if_up.assert_called_once_with(FAKE_INTERFACE, 'network')
        mock_send_member_adv.assert_called_once_with(fixed_ips)

        mock_webob.Response.assert_any_call(
            json={'message': 'OK',
                  'details': f'Updated existing interface {FAKE_INTERFACE}'},
            status=202)

    @mock.patch.object(plug, "webob")
    @mock.patch('octavia.amphorae.backends.agent.api_server.plug.Plug.'
                '_netns_interface_exists', return_value=True)
    @mock.patch('octavia.amphorae.backends.agent.api_server.plug.Plug.'
                '_netns_interface_by_mac', return_value=FAKE_INTERFACE)
    @mock.patch('pyroute2.NetNS', create=True)
    @mock.patch("octavia.amphorae.backends.agent.api_server.osutils."
                "BaseOS.write_vip_interface_file")
    @mock.patch("octavia.amphorae.backends.agent.api_server.osutils."
                "BaseOS.bring_interface_up")
    @mock.patch("octavia.amphorae.backends.agent.api_server.util."
                "send_member_advertisements")
    def test_plug_network_on_vip(
            self, mock_send_member_adv, mock_if_up, mock_write_vip_interface,
            mock_netns, mock_by_mac, mock_interface_exists, mock_webob):
        fixed_ips = [
            {'ip_address': FAKE_IP_IPV4,
             'subnet_cidr': FAKE_CIDR_IPV4,
             'gateway': FAKE_GATEWAY_IPV4,
             'host_routes': [
                 {'destination': '192.0.2.128/25',
                  'nexthop': '192.0.2.100'}]
             }, {'ip_address': FAKE_IP_IPV6,
                 'subnet_cidr': FAKE_CIDR_IPV6,
                 'gateway': FAKE_GATEWAY_IPV6,
                 'host_routes': [
                     {'destination': '2001:db8::/64',
                      'nexthop': '2001:db8::ffff'}]
                 }]
        mtu = 1400
        vip_net_info = {
            'vip': '192.0.2.10',
            'subnet_cidr': '192.0.2.0/25',
            'vrrp_ip': '192.0.2.11',
            'gateway': '192.0.2.1',
            'host_routes': []
        }

        m = mock.mock_open()
        with mock.patch('os.open'), mock.patch.object(os, 'fdopen', m):
            self.test_plug.plug_network(FAKE_MAC_ADDRESS, fixed_ips, mtu=1400,
                                        vip_net_info=vip_net_info)

        mock_write_vip_interface.assert_called_once_with(
            interface=FAKE_INTERFACE,
            vips=[{
                'ip_address': vip_net_info['vip'],
                'ip_version': 4,
                'prefixlen': 25,
                'gateway': vip_net_info['gateway'],
                'host_routes': [],
            }],
            vrrp_info={
                'ip': vip_net_info['vrrp_ip'],
                'ip_version': 4,
                'prefixlen': 25,
                'gateway': vip_net_info['gateway'],
                'host_routes': [],
            },
            fixed_ips=fixed_ips, mtu=mtu, is_sriov=False)

        mock_if_up.assert_called_once_with(FAKE_INTERFACE, 'vip')
        mock_send_member_adv.assert_called_once_with(fixed_ips)

        mock_webob.Response.assert_any_call(
            json={'message': 'OK',
                  'details': f'Updated existing interface {FAKE_INTERFACE}'},
            status=202)

    @mock.patch('pyroute2.NetNS', create=True)
    def test__netns_get_next_interface(self, mock_netns):
        netns_handle = mock_netns.return_value.__enter__.return_value

        netns_handle.get_links.return_value = [
            {'attrs': [['IFLA_IFNAME', 'lo']]},
        ]

        ifname = self.test_plug._netns_get_next_interface()
        self.assertEqual('eth2', ifname)

        netns_handle.get_links.return_value = [
            {'attrs': [['IFLA_IFNAME', 'lo']]},
            {'attrs': [['IFLA_IFNAME', 'eth1']]},
            {'attrs': [['IFLA_IFNAME', 'eth2']]},
            {'attrs': [['IFLA_IFNAME', 'eth3']]},
        ]

        ifname = self.test_plug._netns_get_next_interface()
        self.assertEqual('eth4', ifname)

        netns_handle.get_links.return_value = [
            {'attrs': [['IFLA_IFNAME', 'lo']]},
            {'attrs': [['IFLA_IFNAME', 'eth1']]},
            {'attrs': [['IFLA_IFNAME', 'eth3']]},
        ]

        ifname = self.test_plug._netns_get_next_interface()
        self.assertEqual('eth2', ifname)

        netns_handle.get_links.return_value = [
            {'attrs': [['IFLA_IFNAME', f'eth{idx}']]}
            for idx in range(2, 1000)]

        ifname = self.test_plug._netns_get_next_interface()
        self.assertEqual('eth1000', ifname)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import warnings

import fixtures
from sqlalchemy import exc as sqla_exc


class WarningsFixture(fixtures.Fixture):
    """Filters out warnings during test runs."""

    def setUp(self):
        super().setUp()
        # Make deprecation warnings only happen once to avoid spamming
        warnings.simplefilter('once', DeprecationWarning)

        # Enable deprecation warnings to capture upcoming SQLAlchemy changes
        warnings.filterwarnings(
            'error',
            category=sqla_exc.SADeprecationWarning)

        self.addCleanup(warnings.resetwarnings)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from octavia.common import constants


rules = [

    # API access methods
    #
    # These are the only rules that should be applied to API endpoints.

    policy.RuleDefault(
        name='load-balancer:read',
        check_str='rule:load-balancer:observer_and_owner or '
                  'rule:load-balancer:global_observer or '
                  'rule:load-balancer:member_and_owner or '
                  'rule:load-balancer:admin',
        scope_types=[constants.RBAC_SCOPE_PROJECT]),

    policy.RuleDefault(
        name='load-balancer:read-global',
        check_str='rule:load-balancer:global_observer or '
                  'rule:load-balancer:admin',
        scope_types=[constants.RBAC_SCOPE_PROJECT]),

    policy.RuleDefault(
        name='load-balancer:write',
        check_str='rule:load-balancer:member_and_owner or '
                  'rule:load-balancer:admin',
        scope_types=[constants.RBAC_SCOPE_PROJECT]),

    policy.RuleDefault(
        name='load-balancer:read-quota',
        check_str='rule:load-balancer:observer_and_owner or '
                  'rule:load-balancer:global_observer or '
                  'rule:load-balancer:member_and_owner or '
                  'rule:load-balancer:quota-admin or '
                  'rule:load-balancer:admin',
        scope_types=[constants.RBAC_SCOPE_PROJECT]),

    policy.RuleDefault(
        name='load-balancer:read-quota-global',
        check_str='rule:load-balancer:global_observer or '
                  'rule:load-balancer:quota-admin or '
                  'rule:load-balancer:admin',
        scope_types=[constants.RBAC_SCOPE_PROJECT]),

    policy.RuleDefault(
        name='load-balancer:write-quota',
        check_str='rule:load-balancer:quota-admin or '
                  'rule:load-balancer:admin',
        scope_types=[constants.RBAC_SCOPE_PROJECT]),
]


def list_rules():
    return rules



# Copyright 2019 Rackspace US Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""extend pool for backend CA and CRL

Revision ID: 74aae261694c
Revises: a1f689aecc1d
Create Date: 2019-02-27 09:22:24.779576

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '74aae261694c'
down_revision = 'a1f689aecc1d'


def upgrade():
    op.add_column('pool', sa.Column('ca_tls_certificate_id', sa.String(255),
                                    nullable=True))
    op.add_column('pool', sa.Column('crl_container_id', sa.String(255),
                                    nullable=True))



# Copyright (c) 2018 China Telecom Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""amphora add flavor id

Revision ID: 4f65b4f91c39
Revises: 80dba23a159f
Create Date: 2018-07-16 09:59:07.169894

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '4f65b4f91c39'
down_revision = '80dba23a159f'


def upgrade():
    op.add_column(
        'amphora',
        sa.Column('compute_flavor', sa.String(255), nullable=True)
    )



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

import time

from cryptography import fernet
from oslo_config import cfg
from oslo_log import log as logging
from stevedore import driver as stevedore_driver
from taskflow import retry
from taskflow import task
from taskflow.types import failure
import tenacity

from octavia.amphorae.backends.agent import agent_jinja_cfg
from octavia.common import constants
from octavia.common import exceptions
from octavia.common.jinja.logging import logging_jinja_cfg
from octavia.common.jinja import user_data_jinja_cfg
from octavia.common import utils
from octavia.controller.worker import amphora_rate_limit
from octavia.db import api as db_apis
from octavia.db import repositories as repo

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class BaseComputeTask(task.Task):
    """Base task to load drivers common to the tasks."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.compute = stevedore_driver.DriverManager(
            namespace='octavia.compute.drivers',
            name=CONF.controller_worker.compute_driver,
            invoke_on_load=True
        ).driver
        self.loadbalancer_repo = repo.LoadBalancerRepository()
        self.rate_limit = amphora_rate_limit.AmphoraBuildRateLimit()


class ComputeRetry(retry.Times):

    def on_failure(self, history, *args, **kwargs):
        last_errors = history[-1][1]
        max_retry_attempt = CONF.controller_worker.amp_active_retries
        for task_name, ex_info in last_errors.items():
            if len(history) <= max_retry_attempt:
                # When taskflow persistence is enabled and flow/task state is
                # saved in the backend. If flow(task) is restored(restart of
                # worker,etc) we are getting ex_info as None - we need to RETRY
                # task to check its real state.
                if ex_info is None or ex_info._exc_info is None:
                    return retry.RETRY
                excp = ex_info._exc_info[1]
                if isinstance(excp, exceptions.ComputeWaitTimeoutException):
                    return retry.RETRY

        return retry.REVERT_ALL


class ComputeCreate(BaseComputeTask):
    """Create the compute instance for a new amphora."""

    def execute(self, amphora_id, server_group_id, config_drive_files=None,
                build_type_priority=constants.LB_CREATE_NORMAL_PRIORITY,
                ports=None, flavor=None, availability_zone=None):
        """Create an amphora

        :param availability_zone: availability zone metadata dictionary

        :returns: an amphora
        """
        ports = ports or []
        network_ids = CONF.controller_worker.amp_boot_network_list[:]
        config_drive_files = config_drive_files or {}
        user_data = None
        LOG.debug("Compute create execute for amphora with id %s", amphora_id)

        user_data_config_drive = CONF.controller_worker.user_data_config_drive
        key_name = CONF.controller_worker.amp_ssh_key_name

        # Apply an Octavia flavor customizations
        if flavor:
            topology = flavor.get(constants.LOADBALANCER_TOPOLOGY,
                                  CONF.controller_worker.loadbalancer_topology)
            amp_compute_flavor = flavor.get(
                constants.COMPUTE_FLAVOR, CONF.controller_worker.amp_flavor_id)
            amp_image_tag = flavor.get(
                constants.AMP_IMAGE_TAG, CONF.controller_worker.amp_image_tag)
        else:
            topology = CONF.controller_worker.loadbalancer_topology
            amp_compute_flavor = CONF.controller_worker.amp_flavor_id
            amp_image_tag = CONF.controller_worker.amp_image_tag

        if availability_zone:
            amp_availability_zone = availability_zone.get(
                constants.COMPUTE_ZONE)
            amp_network = availability_zone.get(constants.MANAGEMENT_NETWORK)
            if amp_network:
                network_ids = [amp_network]
        else:
            amp_availability_zone = None
        try:
            if CONF.haproxy_amphora.build_rate_limit != -1:
                self.rate_limit.add_to_build_request_queue(
                    amphora_id, build_type_priority)

            agent_cfg = agent_jinja_cfg.AgentJinjaTemplater()
            config_drive_files['/etc/octavia/amphora-agent.conf'] = (
                agent_cfg.build_agent_config(amphora_id, topology))

            logging_cfg = logging_jinja_cfg.LoggingJinjaTemplater(
                CONF.amphora_agent.logging_template_override)
            config_drive_files['/etc/rsyslog.d/10-rsyslog.conf'] = (
                logging_cfg.build_logging_config())

            udtemplater = user_data_jinja_cfg.UserDataJinjaCfg()
            user_data = udtemplater.build_user_data_config(
                config_drive_files if user_data_config_drive else {})
            if user_data_config_drive:
                config_drive_files = None

            compute_id = self.compute.build(
                name="amphora-" + amphora_id,
                amphora_flavor=amp_compute_flavor,
                image_tag=amp_image_tag,
                image_owner=CONF.controller_worker.amp_image_owner_id,
                key_name=key_name,
                sec_groups=CONF.controller_worker.amp_secgroup_list,
                network_ids=network_ids,
                port_ids=[port.id for port in ports],
                config_drive_files=config_drive_files,
                user_data=user_data,
                server_group_id=server_group_id,
                availability_zone=amp_availability_zone)

            LOG.info("Server created with id: %s for amphora id: %s",
                     compute_id, amphora_id)
            return compute_id

        except Exception:
            LOG.exception("Compute create for amphora id: %s failed",
                          amphora_id)
            raise

    def revert(self, result, amphora_id, *args, **kwargs):
        """This method will revert the creation of the

        amphora. So it will just delete it in this flow
        """
        if isinstance(result, failure.Failure):
            return
        compute_id = result
        LOG.warning("Reverting compute create for amphora with id "
                    "%(amp)s and compute id: %(comp)s",
                    {'amp': amphora_id, 'comp': compute_id})
        try:
            self.compute.delete(compute_id)
        except Exception:
            LOG.exception("Reverting compute create failed")


class CertComputeCreate(ComputeCreate):
    def execute(self, amphora_id, server_pem, server_group_id,
                build_type_priority=constants.LB_CREATE_NORMAL_PRIORITY,
                ports=None, flavor=None, availability_zone=None):
        """Create an amphora

        :param availability_zone: availability zone metadata dictionary

        :returns: an amphora
        """

        # load client certificate
        with open(CONF.controller_worker.client_ca,
                  encoding='utf-8') as client_ca:
            ca = client_ca.read()

        key = utils.get_compatible_server_certs_key_passphrase()
        fer = fernet.Fernet(key)
        config_drive_files = {
            '/etc/octavia/certs/server.pem': fer.decrypt(
                server_pem.encode("utf-8")).decode("utf-8"),
            '/etc/octavia/certs/client_ca.pem': ca}
        return super().execute(
            amphora_id, config_drive_files=config_drive_files,
            build_type_priority=build_type_priority,
            server_group_id=server_group_id, ports=ports, flavor=flavor,
            availability_zone=availability_zone)


class DeleteAmphoraeOnLoadBalancer(BaseComputeTask):
    """Delete the amphorae on a load balancer.

    Iterate through amphorae, deleting them
    """

    def execute(self, loadbalancer):
        session = db_apis.get_session()
        with session.begin():
            db_lb = self.loadbalancer_repo.get(
                session, id=loadbalancer[constants.LOADBALANCER_ID])
        for amp in db_lb.amphorae:
            # The compute driver will already handle NotFound
            try:
                self.compute.delete(amp.compute_id)
            except Exception:
                LOG.exception("Compute delete for amphora id: %s failed",
                              amp.id)
                raise


class ComputeDelete(BaseComputeTask):
    @tenacity.retry(retry=tenacity.retry_if_exception_type(),
                    stop=tenacity.stop_after_attempt(CONF.compute.max_retries),
                    wait=tenacity.wait_exponential(
                        multiplier=CONF.compute.retry_backoff,
                        min=CONF.compute.retry_interval,
                        max=CONF.compute.retry_max), reraise=True)
    def execute(self, amphora, passive_failure=False):

        amphora_id = amphora.get(constants.ID)
        compute_id = amphora[constants.COMPUTE_ID]

        # tenacity 8.5.0 moves statistics from the retry object to the function
        try:
            retry_statistics = self.execute.statistics
        except AttributeError:
            retry_statistics = self.execute.retry.statistics
        if retry_statistics.get(constants.ATTEMPT_NUMBER, 1) == 1:
            LOG.debug('Compute delete execute for amphora with ID %s and '
                      'compute ID: %s', amphora_id, compute_id)
        else:
            LOG.warning('Retrying compute delete of %s attempt %s of %s.',
                        compute_id,
                        retry_statistics[constants.ATTEMPT_NUMBER],
                        self.execute.retry.stop.max_attempt_number)
        # Let the Taskflow engine know we are working and alive
        # Don't use get with a default for 'attempt_number', we need to fail
        # if that number is missing.
        self.update_progress(
            retry_statistics[constants.ATTEMPT_NUMBER] /
            self.execute.retry.stop.max_attempt_number)

        try:
            self.compute.delete(compute_id)
        except Exception:
            if (retry_statistics[constants.ATTEMPT_NUMBER] !=
                    self.execute.retry.stop.max_attempt_number):
                LOG.warning('Compute delete for amphora id: %s failed. '
                            'Retrying.', amphora_id)
                raise
            if passive_failure:
                LOG.exception('Compute delete for compute ID: %s on amphora '
                              'ID: %s failed. This resource will be abandoned '
                              'and should manually be cleaned up once the '
                              'compute service is functional.',
                              compute_id, amphora_id)
            else:
                LOG.exception('Compute delete for compute ID: %s on amphora '
                              'ID: %s failed. The compute service has failed. '
                              'Aborting and reverting.', compute_id,
                              amphora_id)
                raise


class ComputeWait(BaseComputeTask):
    """Wait for the compute driver to mark the amphora active."""

    def execute(self, compute_id, amphora_id, availability_zone):
        """Wait for the compute driver to mark the amphora active

        :param compute_id: virtual machine UUID
        :param amphora_id: id of the amphora object
        :param availability_zone: availability zone metadata dictionary

        :raises: Generic exception if the amphora is not active
        :returns: An amphora object
        """
        if availability_zone:
            amp_network = availability_zone.get(constants.MANAGEMENT_NETWORK)
        else:
            amp_network = None

        amp, fault = self.compute.get_amphora(compute_id, amp_network)
        if amp.status == constants.ACTIVE:
            if CONF.haproxy_amphora.build_rate_limit != -1:
                self.rate_limit.remove_from_build_req_queue(amphora_id)
            return amp.to_dict()
        if amp.status == constants.ERROR:
            raise exceptions.ComputeBuildException(fault=fault)

        time.sleep(CONF.controller_worker.amp_active_wait_sec)
        raise exceptions.ComputeWaitTimeoutException(id=compute_id)


class NovaServerGroupCreate(BaseComputeTask):
    def execute(self, loadbalancer_id):
        """Create a server group by nova client api

        :param loadbalancer_id: will be used for server group's name
        :param policy: will used for server group's policy
        :raises: Generic exception if the server group is not created
        :returns: server group's id
        """

        name = 'octavia-lb-' + loadbalancer_id
        server_group = self.compute.create_server_group(
            name, CONF.nova.anti_affinity_policy)
        LOG.debug("Server Group created with id: %s for load balancer id: "
                  "%s", server_group.id, loadbalancer_id)
        return server_group.id

    def revert(self, result, *args, **kwargs):
        """This method will revert the creation of the

        :param result: here it refers to server group id
        """
        server_group_id = result
        LOG.warning("Reverting server group create with id:%s",
                    server_group_id)
        try:
            self.compute.delete_server_group(server_group_id)
        except Exception as e:
            LOG.error("Failed to delete server group.  Resources may "
                      "still be in use for server group: %(sg)s due to "
                      "error: %(except)s",
                      {'sg': server_group_id, 'except': str(e)})


class NovaServerGroupDelete(BaseComputeTask):
    def execute(self, server_group_id):
        if server_group_id is not None:
            self.compute.delete_server_group(server_group_id)


class AttachPort(BaseComputeTask):
    def execute(self, amphora, port):
        """Attach a port to an amphora instance.

        :param amphora: The amphora to attach the port to.
        :param port: The port to attach to the amphora.
        :returns: None
        """
        LOG.debug('Attaching port: %s to compute: %s',
                  port[constants.ID], amphora[constants.COMPUTE_ID])
        self.compute.attach_network_or_port(amphora[constants.COMPUTE_ID],
                                            port_id=port[constants.ID])

    def revert(self, amphora, port, *args, **kwargs):
        """Revert our port attach.

        :param amphora: The amphora to detach the port from.
        :param port: The port to attach to the amphora.
        """
        LOG.warning('Reverting port: %s attach to compute: %s',
                    port[constants.ID], amphora[constants.COMPUTE_ID])
        try:
            self.compute.detach_port(amphora[constants.COMPUTE_ID],
                                     port[constants.ID])
        except Exception as e:
            LOG.error('Failed to detach port %s from compute %s for revert '
                      'due to %s.', port[constants.ID],
                      amphora[constants.COMPUTE_ID], str(e))



#    Copyright 2014 Rackspace
#    Copyright 2016 Blue Box, an IBM Company
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from octavia_lib.api.drivers import data_models as driver_dm
from oslo_db import exception as odb_exceptions
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import strutils
from pecan import request as pecan_request
from sqlalchemy.orm import exc as sa_exception
from wsme import types as wtypes
from wsmeext import pecan as wsme_pecan

from octavia.api.drivers import driver_factory
from octavia.api.drivers import utils as driver_utils
from octavia.api.v2.controllers import base
from octavia.api.v2.types import member as member_types
from octavia.common import constants
from octavia.common import data_models
from octavia.common import exceptions
from octavia.common import validate
from octavia.db import prepare as db_prepare
from octavia.i18n import _


LOG = logging.getLogger(__name__)


class MemberController(base.BaseController):
    RBAC_TYPE = constants.RBAC_MEMBER

    def __init__(self, pool_id):
        super().__init__()
        self.pool_id = pool_id

    @wsme_pecan.wsexpose(member_types.MemberRootResponse, wtypes.text,
                         [wtypes.text], ignore_extra_args=True)
    def get(self, id, fields=None):
        """Gets a single pool member's details."""
        context = pecan_request.context.get('octavia_context')
        with context.session.begin():
            db_member = self._get_db_member(context.session, id,
                                            show_deleted=False)

        self._auth_validate_action(context, db_member.project_id,
                                   constants.RBAC_GET_ONE)

        self._validate_pool_id(id, db_member.pool_id)

        result = self._convert_db_to_type(
            db_member, member_types.MemberResponse)
        if fields is not None:
            result = self._filter_fields([result], fields)[0]
        return member_types.MemberRootResponse(member=result)

    @wsme_pecan.wsexpose(member_types.MembersRootResponse, [wtypes.text],
                         ignore_extra_args=True)
    def get_all(self, fields=None):
        """Lists all pool members of a pool."""
        pcontext = pecan_request.context
        context = pcontext.get('octavia_context')

        with context.session.begin():
            pool = self._get_db_pool(context.session, self.pool_id,
                                     show_deleted=False, limited_graph=True)

            self._auth_validate_action(context, pool.project_id,
                                       constants.RBAC_GET_ALL)

            db_members, links = self.repositories.member.get_all_API_list(
                context.session, show_deleted=False,
                pool_id=self.pool_id,
                pagination_helper=pcontext.get(constants.PAGINATION_HELPER),
                limited_graph=True)
        result = self._convert_db_to_type(
            db_members, [member_types.MemberResponse])
        if fields is not None:
            result = self._filter_fields(result, fields)
        return member_types.MembersRootResponse(
            members=result, members_links=links)

    def _get_affected_listener_ids(self, session, member=None):
        """Gets a list of all listeners this request potentially affects."""
        if member:
            listener_ids = [li.id for li in member.pool.listeners]
        else:
            pool = self._get_db_pool(session, self.pool_id)
            listener_ids = [li.id for li in pool.listeners]
        return listener_ids

    def _test_lb_and_listener_and_pool_statuses(self, session, member=None):
        """Verify load balancer is in a mutable state."""
        # We need to verify that any listeners referencing this member's
        # pool are also mutable
        pool = self._get_db_pool(session, self.pool_id)
        # Check the parent is not locked for some reason (ERROR, etc.)
        if pool.provisioning_status not in constants.MUTABLE_STATUSES:
            raise exceptions.ImmutableObject(resource='Pool', id=self.pool_id)
        load_balancer_id = pool.load_balancer_id
        if not self.repositories.test_and_set_lb_and_listeners_prov_status(
                session, load_balancer_id,
                constants.PENDING_UPDATE, constants.PENDING_UPDATE,
                listener_ids=self._get_affected_listener_ids(session, member),
                pool_id=self.pool_id):
            LOG.info("Member cannot be created or modified because the "
                     "Load Balancer is in an immutable state")
            raise exceptions.ImmutableObject(resource='Load Balancer',
                                             id=load_balancer_id)

    def _validate_create_member(self, lock_session, member_dict):
        """Validate creating member on pool."""
        try:
            ret = self.repositories.member.create(lock_session, **member_dict)
            lock_session.flush()
            return ret
        except odb_exceptions.DBDuplicateEntry as e:
            raise exceptions.DuplicateMemberEntry(
                ip_address=member_dict.get('ip_address'),
                port=member_dict.get('protocol_port')) from e
        except odb_exceptions.DBReferenceError as e:
            raise exceptions.InvalidOption(value=member_dict.get(e.key),
                                           option=e.key) from e
        except odb_exceptions.DBError as e:
            raise exceptions.APIException() from e
        return None

    def _validate_pool_id(self, member_id, db_member_pool_id):
        if db_member_pool_id != self.pool_id:
            raise exceptions.NotFound(resource='Member', id=member_id)

    @wsme_pecan.wsexpose(member_types.MemberRootResponse,
                         body=member_types.MemberRootPOST, status_code=201)
    def post(self, member_):
        """Creates a pool member on a pool."""
        member = member_.member
        context = pecan_request.context.get('octavia_context')

        flavor_dict = {}
        with context.session.begin():
            pool = self.repositories.pool.get(context.session, id=self.pool_id)
            member.project_id, provider = self._get_lb_project_id_provider(
                context.session, pool.load_balancer_id)
            if pool.load_balancer.flavor_id:
                try:
                    flavor_dict = (
                        self.repositories.flavor.get_flavor_metadata_dict(
                            context.session, pool.load_balancer.flavor_id))
                except sa_exception.NoResultFound:
                    LOG.error("load balancer has a flavor ID: %s that was not "
                              "found in the database. Assuming no flavor.",
                              pool.load_balancer.flavor_id)

        self._auth_validate_action(context, member.project_id,
                                   constants.RBAC_POST)

        validate.ip_not_reserved(member.address)

        # Validate member subnet
        if (member.subnet_id and
                not validate.subnet_exists(member.subnet_id, context=context)):
            raise exceptions.NotFound(resource='Subnet', id=member.subnet_id)

        # Load the driver early as it also provides validation
        driver = driver_factory.get_driver(provider)

        context.session.begin()
        try:
            if self.repositories.check_quota_met(
                    context.session,
                    data_models.Member,
                    member.project_id):
                raise exceptions.QuotaException(
                    resource=data_models.Member._name())

            db_member_dict = member.to_dict(render_unsets=True)

            # Validate and store port SR-IOV vnic_type
            request_sriov = db_member_dict.pop('request_sriov')
            if (request_sriov and not
                    flavor_dict.get(constants.ALLOW_MEMBER_SRIOV, False)):
                raise exceptions.MemberSRIOVDisabled
            if request_sriov:
                db_member_dict[constants.VNIC_TYPE] = (
                    constants.VNIC_TYPE_DIRECT)
            else:
                db_member_dict[constants.VNIC_TYPE] = (
                    constants.VNIC_TYPE_NORMAL)

            member_dict = db_prepare.create_member(db_member_dict,
                                                   self.pool_id,
                                                   bool(pool.health_monitor))

            self._test_lb_and_listener_and_pool_statuses(context.session)

            db_member = self._validate_create_member(context.session,
                                                     member_dict)

            # Prepare the data for the driver data model
            provider_member = (
                driver_utils.db_member_to_provider_member(db_member))

            # Dispatch to the driver
            LOG.info("Sending create Member %s to provider %s",
                     db_member.id, driver.name)
            driver_utils.call_provider(
                driver.name, driver.member_create, provider_member)

            context.session.commit()
        except Exception:
            with excutils.save_and_reraise_exception():
                context.session.rollback()

        with context.session.begin():
            db_member = self._get_db_member(context.session, db_member.id)
        result = self._convert_db_to_type(db_member,
                                          member_types.MemberResponse)
        return member_types.MemberRootResponse(member=result)

    def _graph_create(self, lock_session, member_dict):
        pool = self.repositories.pool.get(lock_session, id=self.pool_id)

        # Validate and store port SR-IOV vnic_type
        request_sriov = member_dict.pop('request_sriov')
        flavor_dict = {}
        if pool.load_balancer.flavor_id:
            try:
                flavor_dict = (
                    self.repositories.flavor.get_flavor_metadata_dict(
                        lock_session, pool.load_balancer.flavor_id))
            except sa_exception.NoResultFound:
                LOG.error("load balancer has a flavor ID: %s that was not "
                          "found in the database. Assuming no flavor.",
                          pool.load_balancer.flavor_id)
        if (request_sriov and not
                flavor_dict.get(constants.ALLOW_MEMBER_SRIOV, False)):
            raise exceptions.MemberSRIOVDisabled

        if request_sriov:
            member_dict[constants.VNIC_TYPE] = constants.VNIC_TYPE_DIRECT
        else:
            member_dict[constants.VNIC_TYPE] = constants.VNIC_TYPE_NORMAL

        member_dict = db_prepare.create_member(
            member_dict, self.pool_id, bool(pool.health_monitor))
        db_member = self._validate_create_member(lock_session, member_dict)

        return db_member

    def _set_default_on_none(self, member):
        """Reset settings to their default values if None/null was passed in

        A None/null value can be passed in to clear a value. PUT values
        that were not provided by the user have a type of wtypes.UnsetType.
        If the user is attempting to clear values, they should either
        be set to None (for example in the name field) or they should be
        reset to their default values.
        This method is intended to handle those values that need to be set
        back to a default value.
        """
        if member.backup is None:
            member.backup = False
        if member.weight is None:
            member.weight = constants.DEFAULT_WEIGHT

    @wsme_pecan.wsexpose(member_types.MemberRootResponse,
                         wtypes.text, body=member_types.MemberRootPUT,
                         status_code=200)
    def put(self, id, member_):
        """Updates a pool member."""
        member = member_.member
        context = pecan_request.context.get('octavia_context')
        with context.session.begin():
            db_member = self._get_db_member(context.session, id,
                                            show_deleted=False)
            pool = self.repositories.pool.get(context.session,
                                              id=db_member.pool_id)
            project_id, provider = self._get_lb_project_id_provider(
                context.session, pool.load_balancer_id)

        self._auth_validate_action(context, project_id, constants.RBAC_PUT)

        self._validate_pool_id(id, db_member.pool_id)

        self._set_default_on_none(member)

        # Load the driver early as it also provides validation
        driver = driver_factory.get_driver(provider)

        with context.session.begin():
            self._test_lb_and_listener_and_pool_statuses(context.session,
                                                         member=db_member)

            # Prepare the data for the driver data model
            member_dict = member.to_dict(render_unsets=False)
            member_dict['id'] = id
            provider_member_dict = (
                driver_utils.member_dict_to_provider_dict(member_dict))

            # Also prepare the baseline object data
            old_provider_member = driver_utils.db_member_to_provider_member(
                db_member)

            # Dispatch to the driver
            LOG.info("Sending update Member %s to provider %s", id,
                     driver.name)
            driver_utils.call_provider(
                driver.name, driver.member_update,
                old_provider_member,
                driver_dm.Member.from_dict(provider_member_dict))

            # Update the database to reflect what the driver just accepted
            member.provisioning_status = constants.PENDING_UPDATE
            db_member_dict = member.to_dict(render_unsets=False)
            self.repositories.member.update(context.session, id,
                                            **db_member_dict)

        # Force SQL alchemy to query the DB, otherwise we get inconsistent
        # results
        context.session.expire_all()
        with context.session.begin():
            db_member = self._get_db_member(context.session, id)
        result = self._convert_db_to_type(db_member,
                                          member_types.MemberResponse)
        return member_types.MemberRootResponse(member=result)

    @wsme_pecan.wsexpose(None, wtypes.text, status_code=204)
    def delete(self, id):
        """Deletes a pool member."""
        context = pecan_request.context.get('octavia_context')
        with context.session.begin():
            db_member = self._get_db_member(context.session, id,
                                            show_deleted=False)

            pool = self.repositories.pool.get(context.session,
                                              id=db_member.pool_id)
            project_id, provider = self._get_lb_project_id_provider(
                context.session, pool.load_balancer_id)

        self._auth_validate_action(context, project_id, constants.RBAC_DELETE)

        self._validate_pool_id(id, db_member.pool_id)

        # Load the driver early as it also provides validation
        driver = driver_factory.get_driver(provider)

        with context.session.begin():
            self._test_lb_and_listener_and_pool_statuses(context.session,
                                                         member=db_member)
            self.repositories.member.update(
                context.session, db_member.id,
                provisioning_status=constants.PENDING_DELETE)

            LOG.info("Sending delete Member %s to provider %s", id,
                     driver.name)
            provider_member = (
                driver_utils.db_member_to_provider_member(db_member))
            driver_utils.call_provider(driver.name, driver.member_delete,
                                       provider_member)


class MembersController(MemberController):

    def __init__(self, pool_id):
        super().__init__(pool_id)

    @wsme_pecan.wsexpose(None, wtypes.text,
                         body=member_types.MembersRootPUT, status_code=202)
    def put(self, additive_only=False, members_=None):
        """Updates all members."""
        members = members_.members
        additive_only = strutils.bool_from_string(additive_only)
        context = pecan_request.context.get('octavia_context')

        with context.session.begin():
            db_pool = self._get_db_pool(context.session, self.pool_id)

            project_id, provider = self._get_lb_project_id_provider(
                context.session, db_pool.load_balancer_id)

        # Check POST+PUT+DELETE since this operation is all of 'CUD'
        self._auth_validate_action(context, project_id, constants.RBAC_POST)
        self._auth_validate_action(context, project_id, constants.RBAC_PUT)
        if not additive_only:
            self._auth_validate_action(context, project_id,
                                       constants.RBAC_DELETE)

        # Load the driver early as it also provides validation
        driver = driver_factory.get_driver(provider)

        with context.session.begin():
            self._test_lb_and_listener_and_pool_statuses(context.session)

            # Reload the pool, the members may have been updated between the
            # first query in this function and the lock of the loadbalancer
            db_pool = self._get_db_pool(context.session, self.pool_id)
            old_members = db_pool.members

            old_member_uniques = {
                (m.ip_address, m.protocol_port): m.id for m in old_members}
            new_member_uniques = [
                (m.address, m.protocol_port) for m in members]

            # Find members that are brand new or updated
            new_members = []
            updated_members = []
            updated_member_uniques = set()
            for m in members:
                key = (m.address, m.protocol_port)
                if key not in old_member_uniques:
                    validate.ip_not_reserved(m.address)
                    new_members.append(m)
                else:
                    m.id = old_member_uniques[key]
                    if key in updated_member_uniques:
                        LOG.error("Member %s is updated multiple times in "
                                  "the same batch request.", m.id)
                        raise exceptions.ValidationException(
                            detail=_("Member must be updated only once in the "
                                     "same request."))
                    updated_member_uniques.add(key)
                    updated_members.append(m)

            # Find members that are deleted
            deleted_members = []
            for m in old_members:
                if (m.ip_address, m.protocol_port) not in new_member_uniques:
                    deleted_members.append(m)

            if not (deleted_members or new_members or updated_members):
                LOG.info("Member batch update is a noop, rolling back and "
                         "returning early.")
                context.session.rollback()
                return

            if additive_only:
                member_count_diff = len(new_members)
            else:
                member_count_diff = len(new_members) - len(deleted_members)
            if member_count_diff > 0 and self.repositories.check_quota_met(
                    context.session, data_models.Member,
                    db_pool.project_id, count=member_count_diff):
                raise exceptions.QuotaException(
                    resource=data_models.Member._name())

            provider_members = []
            valid_subnets = set()
            # Create new members
            for m in new_members:
                # NOTE(mnaser): In order to avoid hitting the Neutron API hard
                # when creating many new members, we cache the
                # validation results. We also validate new
                # members only since subnet ID is immutable.
                # If the member doesn't have a subnet, or the subnet is
                # already valid, move on. Run validate and add it to
                # cache otherwise.
                if m.subnet_id and m.subnet_id not in valid_subnets:
                    # If the subnet does not exist,
                    # raise an exception and get out.
                    if not validate.subnet_exists(
                            m.subnet_id, context=context):
                        raise exceptions.NotFound(
                            resource='Subnet', id=m.subnet_id)

                    # Mark the subnet as valid for future runs.
                    valid_subnets.add(m.subnet_id)

                m = m.to_dict(render_unsets=False)
                m['project_id'] = db_pool.project_id
                created_member = self._graph_create(context.session, m)
                provider_member = driver_utils.db_member_to_provider_member(
                    created_member)
                provider_members.append(provider_member)
            # Update old members
            for m in updated_members:
                m.provisioning_status = constants.PENDING_UPDATE
                m.project_id = db_pool.project_id
                db_member_dict = m.to_dict(render_unsets=False)
                db_member_dict.pop('id')
                # We don't allow updating the vnic_type
                # TODO(johnsom) Give the user an error once we change the
                #               wsme type for batch member update to not use
                #               the MemberPOST type
                db_member_dict.pop(constants.REQUEST_SRIOV)
                self.repositories.member.update(
                    context.session, m.id, **db_member_dict)

                m.pool_id = self.pool_id
                provider_members.append(
                    driver_utils.db_member_to_provider_member(m))
            # Delete old members
            for m in deleted_members:
                if additive_only:
                    # Members are appended to the dict and their status remains
                    # unchanged, because they are logically "untouched".
                    db_member_dict = m.to_dict(render_unsets=False)
                    db_member_dict.pop('id')
                    m.pool_id = self.pool_id
                    provider_members.append(
                        driver_utils.db_member_to_provider_member(m))
                else:
                    # Members are changed to PENDING_DELETE and not passed.
                    self.repositories.member.update(
                        context.session, m.id,
                        provisioning_status=constants.PENDING_DELETE)

            # Dispatch to the driver
            LOG.info("Sending Pool %s batch member update to provider %s",
                     db_pool.id, driver.name)
            driver_utils.call_provider(
                driver.name, driver.member_batch_update, db_pool.id,
                provider_members)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from octavia.common import constants
from octavia.policies import advanced_rbac

rules = [

    # OpenStack keystone default roles

    # Project scoped Member
    policy.RuleDefault(
        name='project-member',
        check_str='role:member and '
                  'project_id:%(project_id)s',
        scope_types=[constants.RBAC_SCOPE_PROJECT]),

    # Project scoped Reader
    policy.RuleDefault(
        name='project-reader',
        check_str='role:reader and '
                  'project_id:%(project_id)s',
        scope_types=[constants.RBAC_SCOPE_PROJECT]),

    policy.RuleDefault(
        name='context_is_admin',
        check_str='role:admin',
        deprecated_rule=advanced_rbac.deprecated_context_is_admin,
        scope_types=[constants.RBAC_SCOPE_PROJECT]),

    # API access roles
    policy.RuleDefault(
        name='load-balancer:admin',
        check_str='is_admin:True or '
                  'role:admin',
        deprecated_rule=advanced_rbac.deprecated_admin,
        scope_types=[constants.RBAC_SCOPE_PROJECT]),

    # Note: 'is_admin:True' is a policy rule that takes into account the
    # auth_strategy == noauth configuration setting.
    # It is equivalent to 'rule:context_is_admin or {auth_strategy == noauth}'

    policy.RuleDefault(
        name='load-balancer:global_observer',
        check_str='role:admin',
        deprecated_rule=advanced_rbac.deprecated_global_observer,
        scope_types=[constants.RBAC_SCOPE_PROJECT]),

    policy.RuleDefault(
        name='load-balancer:member_and_owner',
        check_str='rule:project-member',
        deprecated_rule=advanced_rbac.deprecated_member_and_owner,
        scope_types=[constants.RBAC_SCOPE_PROJECT]),

    policy.RuleDefault(
        name='load-balancer:observer_and_owner',
        check_str='rule:project-reader',
        deprecated_rule=advanced_rbac.deprecated_observer_and_owner,
        scope_types=[constants.RBAC_SCOPE_PROJECT]),

    policy.RuleDefault(
        name='load-balancer:quota-admin',
        check_str='role:admin',
        deprecated_rule=advanced_rbac.deprecated_quota_admin,
        scope_types=[constants.RBAC_SCOPE_PROJECT]),
]


def list_rules():
    return rules



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import hashlib
import io
import os
import re
import shutil
import stat
import subprocess
import time

import flask
import jinja2
from oslo_config import cfg
from oslo_log import log as logging
import webob
from werkzeug import exceptions

from octavia.amphorae.backends.agent.api_server import haproxy_compatibility
from octavia.amphorae.backends.agent.api_server import util
from octavia.amphorae.backends.utils import haproxy_query
from octavia.common import constants as consts
from octavia.common import utils as octavia_utils

LOG = logging.getLogger(__name__)
BUFFER = 100
HAPROXY_RELOAD_RETRIES = 3
HAPROXY_QUERY_RETRIES = 5

CONF = cfg.CONF

SYSTEMD_CONF = 'systemd.conf.j2'

JINJA_ENV = jinja2.Environment(
    autoescape=True,
    loader=jinja2.FileSystemLoader(os.path.dirname(
        os.path.realpath(__file__)
    ) + consts.AGENT_API_TEMPLATES))
SYSTEMD_TEMPLATE = JINJA_ENV.get_template(SYSTEMD_CONF)


# Wrap a stream so we can compute the md5 while reading
class Wrapped:
    def __init__(self, stream_):
        self.stream = stream_
        self.hash = hashlib.md5(usedforsecurity=False)  # nosec

    def read(self, line):
        block = self.stream.read(line)
        if block:
            self.hash.update(block)
        return block

    def get_md5(self):
        return self.hash.hexdigest()

    def __getattr__(self, attr):
        return getattr(self.stream, attr)


class Loadbalancer:

    def get_haproxy_config(self, lb_id):
        """Gets the haproxy config

        :param listener_id: the id of the listener
        """
        self._check_lb_exists(lb_id)
        with open(util.config_path(lb_id), encoding='utf-8') as file:
            cfg = file.read()
            resp = webob.Response(cfg, content_type='text/plain')
            resp.headers['ETag'] = (
                hashlib.md5(octavia_utils.b(cfg),
                            usedforsecurity=False).hexdigest())  # nosec
            return resp

    def upload_haproxy_config(self, amphora_id, lb_id):
        """Upload the haproxy config

        :param amphora_id: The id of the amphora to update
        :param lb_id: The id of the loadbalancer
        """
        stream = Wrapped(flask.request.stream)
        # We have to hash here because HAProxy has a string length limitation
        # in the configuration file "peer <peername>" lines
        peer_name = octavia_utils.base64_sha1_string(amphora_id).rstrip('=')
        if not os.path.exists(util.haproxy_dir(lb_id)):
            os.makedirs(util.haproxy_dir(lb_id))

        name = os.path.join(util.haproxy_dir(lb_id), 'haproxy.cfg.new')
        flags = os.O_WRONLY | os.O_CREAT | os.O_TRUNC
        # mode 00600
        mode = stat.S_IRUSR | stat.S_IWUSR
        b = stream.read(BUFFER)
        s_io = io.StringIO()
        while b:
            # Write haproxy configuration to StringIO
            s_io.write(b.decode('utf8'))
            b = stream.read(BUFFER)

        # Since haproxy user_group is now auto-detected by the amphora agent,
        # remove it from haproxy configuration in case it was provided
        # by an older Octavia controller. This is needed in order to prevent
        # a duplicate entry for 'group' in haproxy configuration, which will
        # result an error when haproxy starts.
        new_config = re.sub(r"\s+group\s.+", "", s_io.getvalue())

        # Handle any haproxy version compatibility issues
        new_config = haproxy_compatibility.process_cfg_for_version_compat(
            new_config)

        with os.fdopen(os.open(name, flags, mode), 'w') as file:
            file.write(new_config)

        # use haproxy to check the config
        cmd = (f"haproxy -c -L {peer_name} -f {name} -f "
               f"{consts.HAPROXY_USER_GROUP_CFG}")

        try:
            subprocess.check_output(cmd.split(), stderr=subprocess.STDOUT,
                                    encoding='utf-8')
        except subprocess.CalledProcessError as e:
            LOG.error("Failed to verify haproxy file: %s %s", e, e.output)
            # Save the last config that failed validation for debugging
            os.rename(name, ''.join([name, '-failed']))
            return webob.Response(
                json={'message': "Invalid request", 'details': e.output},
                status=400)

        # file ok - move it
        os.rename(name, util.config_path(lb_id))

        init_path = util.init_path(lb_id)

        template = SYSTEMD_TEMPLATE
        # Render and install the network namespace systemd service
        util.install_netns_systemd_service()
        util.run_systemctl_command(
            consts.ENABLE, consts.AMP_NETNS_SVC_PREFIX + '.service', False)

        # mode 00644
        mode = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH

        hap_major, hap_minor = haproxy_compatibility.get_haproxy_versions()
        if not os.path.exists(init_path):
            with os.fdopen(os.open(init_path, flags, mode), 'w') as text_file:

                text = template.render(
                    peer_name=peer_name,
                    haproxy_pid=util.pid_path(lb_id),
                    haproxy_cmd=util.CONF.haproxy_amphora.haproxy_cmd,
                    haproxy_cfg=util.config_path(lb_id),
                    haproxy_state_file=util.state_file_path(lb_id),
                    haproxy_socket=util.haproxy_sock_path(lb_id),
                    haproxy_user_group_cfg=consts.HAPROXY_USER_GROUP_CFG,
                    amphora_netns=consts.AMP_NETNS_SVC_PREFIX,
                    amphora_nsname=consts.AMPHORA_NAMESPACE,
                    haproxy_major_version=hap_major,
                    haproxy_minor_version=hap_minor
                )
                text_file.write(text)

        # Make sure the new service is enabled on boot
        try:
            util.run_systemctl_command(
                consts.ENABLE, consts.LOADBALANCER_SYSTEMD % lb_id)
        except subprocess.CalledProcessError as e:
            return webob.Response(json={
                'message': "Error enabling octavia-keepalived service",
                'details': e.output}, status=500)

        res = webob.Response(json={'message': 'OK'}, status=202)
        res.headers['ETag'] = stream.get_md5()

        return res

    def _check_haproxy_uptime(self, lb_id):
        stat_sock_file = util.haproxy_sock_path(lb_id)
        lb_query = haproxy_query.HAProxyQuery(stat_sock_file)
        retries = HAPROXY_QUERY_RETRIES
        for idx in range(retries):
            try:
                info = lb_query.show_info()
                uptime_sec = info['Uptime_sec']
            except Exception as e:
                LOG.warning('Failed to get haproxy info: %s, retrying.', e)
                time.sleep(1)
                continue
            uptime = int(uptime_sec)
            return uptime
        LOG.error('Failed to get haproxy uptime after %d tries.', retries)
        return None

    def start_stop_lb(self, lb_id, action):
        action = action.lower()
        if action not in [consts.AMP_ACTION_START,
                          consts.AMP_ACTION_STOP,
                          consts.AMP_ACTION_RELOAD]:
            return webob.Response(json={
                'message': 'Invalid Request',
                'details': f"Unknown action: {action}"}, status=400)

        self._check_lb_exists(lb_id)
        is_vrrp = (CONF.controller_worker.loadbalancer_topology ==
                   consts.TOPOLOGY_ACTIVE_STANDBY)

        if is_vrrp:
            util.vrrp_check_script_update(lb_id, action)

        # HAProxy does not start the process when given a reload
        # so start it if haproxy is not already running
        if action == consts.AMP_ACTION_RELOAD:
            if consts.OFFLINE == self._check_haproxy_status(lb_id):
                action = consts.AMP_ACTION_START
            else:
                # We first have to save the state when we reload
                haproxy_state_file = util.state_file_path(lb_id)
                stat_sock_file = util.haproxy_sock_path(lb_id)

                lb_query = haproxy_query.HAProxyQuery(stat_sock_file)
                if not lb_query.save_state(haproxy_state_file):
                    # We accept to reload haproxy even if the state_file is
                    # not generated, but we probably want to know about that
                    # failure!
                    LOG.warning('Failed to save haproxy-%s state!', lb_id)

        retries = (HAPROXY_RELOAD_RETRIES
                   if action == consts.AMP_ACTION_RELOAD
                   else 1)
        saved_exc = None
        for idx in range(retries):
            try:
                util.run_systemctl_command(
                    action, consts.LOADBALANCER_SYSTEMD % lb_id)
            except subprocess.CalledProcessError as e:
                # Mitigation for
                # https://bugs.launchpad.net/octavia/+bug/2054666
                if ('is not active, cannot reload.' in e.output and
                        action == consts.AMP_ACTION_RELOAD):

                    saved_exc = e

                    # Wait a few seconds and check that haproxy was restarted
                    uptime = self._check_haproxy_uptime(lb_id)
                    # If haproxy is not reachable or was restarted more than 15
                    # sec ago, let's retry (or maybe restart?)
                    if not uptime or uptime > 15:
                        continue
                    # haproxy probably crashed and was restarted, log it and
                    # continue
                    LOG.warning("An error occured with haproxy while it "
                                "was reloaded, check the haproxy logs for "
                                "more details.")
                    break
                if 'Job is already running' not in e.output:
                    return webob.Response(json={
                        'message': f"Error {action}ing haproxy",
                        'details': e.output
                    }, status=500)
            break
        else:
            # no break, we reach the retry limit for reloads
            return webob.Response(json={
                'message': f"Error {action}ing haproxy",
                'details': saved_exc.output}, status=500)

        # If we are not in active/standby we need to send an IP
        # advertisement (GARP or NA). Keepalived handles this for
        # active/standby load balancers.
        if not is_vrrp and action in [consts.AMP_ACTION_START,
                                      consts.AMP_ACTION_RELOAD]:
            util.send_vip_advertisements(lb_id)

        if action in [consts.AMP_ACTION_STOP,
                      consts.AMP_ACTION_RELOAD]:
            return webob.Response(json={
                'message': 'OK',
                'details': f'Listener {lb_id} {action}ed'}, status=202)

        details = (
            f'Configuration file is valid\nhaproxy daemon for {lb_id} started'
        )

        return webob.Response(json={'message': 'OK', 'details': details},
                              status=202)

    def delete_lb(self, lb_id):
        try:
            self._check_lb_exists(lb_id)
        except exceptions.HTTPException:
            return webob.Response(json={'message': 'OK'})

        # check if that haproxy is still running and if stop it
        if os.path.exists(util.pid_path(lb_id)) and os.path.exists(
                os.path.join('/proc', util.get_haproxy_pid(lb_id))):
            try:
                util.run_systemctl_command(
                    consts.STOP, consts.LOADBALANCER_SYSTEMD % lb_id)
            except subprocess.CalledProcessError as e:
                LOG.error("Failed to stop haproxy-%s service: %s %s",
                          lb_id, e, e.output)
                return webob.Response(json={
                    'message': "Error stopping haproxy",
                    'details': e.output}, status=500)

        # parse config and delete stats socket
        try:
            stats_socket = util.parse_haproxy_file(lb_id)[0]
            os.remove(stats_socket)
        except Exception:
            pass

        # Since this script should be deleted at LB delete time
        # we can check for this path to see if VRRP is enabled
        # on this amphora and not write the file if VRRP is not in use
        if os.path.exists(util.keepalived_check_script_path()):
            util.vrrp_check_script_update(
                lb_id, action=consts.AMP_ACTION_STOP)

        # delete the ssl files
        try:
            shutil.rmtree(self._cert_dir(lb_id))
        except Exception:
            pass

        # disable the service
        init_path = util.init_path(lb_id)

        util.run_systemctl_command(
            consts.DISABLE, consts.LOADBALANCER_SYSTEMD % lb_id, False)

        # delete the directory + init script for that listener
        shutil.rmtree(util.haproxy_dir(lb_id))
        if os.path.exists(init_path):
            os.remove(init_path)

        return webob.Response(json={'message': 'OK'})

    def get_all_listeners_status(self, other_listeners=None):
        """Gets the status of all listeners

        This method will not consult the stats socket
        so a listener might show as ACTIVE but still be
        in ERROR

        Currently type==SSL is also not detected
        """
        listeners = []

        for lb in util.get_loadbalancers():
            stats_socket, listeners_on_lb = util.parse_haproxy_file(lb)

            for listener_id, listener in listeners_on_lb.items():
                listeners.append({
                    'status': consts.ACTIVE,
                    'uuid': listener_id,
                    'type': listener['mode'],
                })

        if other_listeners:
            listeners = listeners + other_listeners
        return webob.Response(json=listeners, content_type='application/json')

    def upload_certificate(self, lb_id, filename):
        self._check_ssl_filename_format(filename)

        # create directory if not already there
        if not os.path.exists(self._cert_dir(lb_id)):
            os.makedirs(self._cert_dir(lb_id))

        stream = Wrapped(flask.request.stream)
        file = self._cert_file_path(lb_id, filename)
        flags = os.O_WRONLY | os.O_CREAT | os.O_TRUNC
        # mode 00600
        mode = stat.S_IRUSR | stat.S_IWUSR
        with os.fdopen(os.open(file, flags, mode), 'wb') as crt_file:
            b = stream.read(BUFFER)
            while b:
                crt_file.write(b)
                b = stream.read(BUFFER)

        resp = webob.Response(json={'message': 'OK'})
        resp.headers['ETag'] = stream.get_md5()
        return resp

    def get_certificate_md5(self, lb_id, filename):
        self._check_ssl_filename_format(filename)

        cert_path = self._cert_file_path(lb_id, filename)
        path_exists = os.path.exists(cert_path)
        if not path_exists:
            return webob.Response(json={
                'message': 'Certificate Not Found',
                'details': f"No certificate with filename: {filename}"},
                status=404)

        with open(cert_path, encoding='utf-8') as crt_file:
            cert = crt_file.read()
            md5sum = hashlib.md5(octavia_utils.b(cert),
                                 usedforsecurity=False).hexdigest()  # nosec
            resp = webob.Response(json={'md5sum': md5sum})
            resp.headers['ETag'] = md5sum
            return resp

    def delete_certificate(self, lb_id, filename):
        self._check_ssl_filename_format(filename)
        if os.path.exists(self._cert_file_path(lb_id, filename)):
            os.remove(self._cert_file_path(lb_id, filename))
        return webob.Response(json={'message': 'OK'})

    def _get_listeners_on_lb(self, lb_id):
        if os.path.exists(util.pid_path(lb_id)):
            if os.path.exists(
                    os.path.join('/proc', util.get_haproxy_pid(lb_id))):
                # Check if the listener is disabled
                with open(util.config_path(lb_id), encoding='utf-8') as file:
                    cfg = file.read()
                    m = re.findall('^frontend (.*)$', cfg, re.MULTILINE)
                    return m or []
            else:  # pid file but no process...
                return []
        else:
            return []

    def _check_lb_exists(self, lb_id):
        # check if we know about that lb
        if lb_id not in util.get_loadbalancers():
            raise exceptions.HTTPException(
                response=webob.Response(json={
                    'message': 'Loadbalancer Not Found',
                    'details': f"No loadbalancer with UUID: {lb_id}"},
                    status=404))

    def _check_ssl_filename_format(self, filename):
        # check if the format is (xxx.)*xxx.pem
        if not re.search(r'(\w.)+pem', filename):
            raise exceptions.HTTPException(
                response=webob.Response(json={
                    'message': 'Filename has wrong format'}, status=400))

    def _cert_dir(self, lb_id):
        return os.path.join(util.CONF.haproxy_amphora.base_cert_dir, lb_id)

    def _cert_file_path(self, lb_id, filename):
        return os.path.join(self._cert_dir(lb_id), filename)

    def _check_haproxy_status(self, lb_id):
        if os.path.exists(util.pid_path(lb_id)):
            if os.path.exists(
                    os.path.join('/proc', util.get_haproxy_pid(lb_id))):
                return consts.ACTIVE
        return consts.OFFLINE
